"""
This file contains the zayocli main entrypoint for CLI purposes.  The `setup.py`
uses this file to install the CLI tool during package installation.

This file also servers to import each of the CLI components, which is why the
import modules list below exist.
"""
from .cli_root import cli

from . import cli_mtc  # noqa
from . import cli_svcinv  # noqa


def main():
    """ CLI main entrypoint """
    cli()



from pyzayo.mtc_mixin import ZayoMatenanceMixin
from pyzayo.svcinv_mixin import ZayoServiceInventoryMixin


__all__ = ["ZayoClient"]


class ZayoClient(ZayoMatenanceMixin, ZayoServiceInventoryMixin):
    """
    Zayo Client class supporting the Maintenance and Sevice-Inventory functional areas.
    """

    pass



"""
This file contains _constants_ used by this package.
"""

from enum import Enum

# Environment variables

Env = {"client_id": "ZAYO_CLIENT_ID", "client_secret": "ZAYO_CLIENT_SECRET"}

# URL for authorizing client crendentials and obtaining an access token
ZAYO_URL_AUTH = "https://auth.testzayo.com/oauth/token"

# WARNING: base URL must end with "/"
# Reference: https://github.com/encode/httpx/issues/846

ZAYO_URL_SM = "https://api.zayo.com/services/service-management/v1/"

# WARNING: path routes from base_url MUST NOT begin with a "/" for the same resaons.
ZAYO_SM_ROUTE_MTC_CASES = "maintenance-cases"
ZAYO_SM_ROUTE_MTC_IMPACTS = "maintenance-impacts"
ZAYO_SM_ROUTE_MTC_NOTIFS_BY_CASE = "maintenance-cases/{case_num}/notifications"
ZAYO_SM_ROUTE_MTC_NOTIFS_BY_NAME = "maintenance-cases/notifications/{name}"
ZAYO_SM_ROUTE_SERVICES = "existing-services"


# The Zayo API has a maximum "top" count of 50 records
MAX_TOP_COUNT = 50
MAX_PAGED_RECORDS = 100


class NotificationTypes(Enum):
    """
    Used by the notification API endpoints
    """

    scheduled = "Scheduled"
    rescheduled = "Rescheduled"
    started = "Maintenance Started"
    stopped = "Maintenance Stopped"
    completed = "Maintenance Completed"
    cancelled = "Cancelled"


class OrderBy(Enum):
    """
    Used for any API supporting the `orderBy` criteria
    """

    date_later = "primaryDate desc"
    date_sooner = "primaryDate asc"


class CaseUrgencyOptions(str, Enum):
    """
    Used by the Cases `urgency` field
    """

    planned = "Planned"
    emergency = "Emergency"
    demand = "Demand"


class CaseStatusOptions(str, Enum):
    """
    Used by case record `status` field
    """

    new = "New"
    in_review = "MR Review"
    rejected = "MR Rejected"
    impact_fe_in_review = "Impact FE Review"
    impact_fe_complete = "Impact FE Complete"
    pending_schedule = "MR Pending Schedule"
    schedule_in_progress = "Scheduling in Progress"
    started = "Maint Started"
    stopped = "Maint Stopped"
    completed = "Maint Completed"
    rescheduled = "Rescheduled"
    cancelled = "Cancelled"
    closed = "Closed"
    on_hold_bcdr = "On-Hold BCDR"
    on_hold = "On-Hold"
    scheduled = "Scheduled"
    COVID = "COVID"


class CaseImpactOptions(Enum):
    """
    Used by Case record impact field.
    """

    potential_svc_aff = "Potential Service Affecting"
    svc_aff = "Service Affecting"


REQ_MOST_RECENT = {"paging": {"top": 1}, "orderBy": [OrderBy.date_later.value]}
REQ_OLDEST = {"paging": {"top": 1}, "orderBy": [OrderBy.date_sooner.value]}


class InventoryStatusOption(str, Enum):
    """ serivce inventory.status field options """

    active = "Active"
    pending = "Pending Change"



from typing import Optional
from datetime import time, date, datetime

from pydantic import BaseModel, Field

from pyzayo import consts

__all__ = ["CaseRecord", "ImpactRecord", "NotificationDetailRecord"]


class CaseRecord(BaseModel):
    """
    This model defines a maintencance case record returned via the API
    """

    case_id: str = Field(alias="caseId")
    case_num: str = Field(alias="caseNumber")
    urgency: str = Field(alias="urgency")
    impact: str = Field(alias="levelOfImpact")
    status: consts.CaseStatusOptions
    primary_date: date = Field(alias="primaryDate")
    primary_date_2: Optional[date] = Field(alias="x2ndPrimaryDate")
    primary_date_3: Optional[date] = Field(alias="x3rdPrimaryDate")
    from_time: time = Field(alias="fromTime")
    to_time: time = Field(alias="toTime")
    reason: str = Field(alias="reasonForMaintenance")
    location: str
    longitiude: Optional[float]
    latittude: Optional[float]


class ImpactRecord(BaseModel):
    """
    This model defines fields in the maintenance impact record (of interest, not all)
    """

    case_num: str = Field(alias="caseNumber")
    circuit_id: str = Field(alias="circuitId")
    impact: str = Field(alias="expectedImpact")
    clli_a: str = Field(alias="aLocationClli")
    clli_z: str = Field(alias="zLocationClli")


class NotificationDetailRecord(BaseModel):
    """
    This model defines the fields in the maintenance notification details record
    """

    name: str
    type: str = Field(alias="notificationType")
    date: datetime = Field(alias="lastModifiedDate")
    subject: str
    email_list: str = Field(alias="toEmailList")
    email_content: str = Field(alias="emailBody")



"""
This file contains the CLI command for the ZAYO API service inventory feature area.

References
----------
    For the Rich package, colors are defined here:
    https://rich.readthedocs.io/en/latest/appendix/colors.html#appendix-colors

"""

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import List, Dict

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

from rich.console import Console
from rich.table import Table, Text

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from pyzayo import ZayoClient
from .cli_root import cli
from pyzayo.consts import InventoryStatusOption

# -----------------------------------------------------------------------------
#
#                               TABLE CODE BEGINS
#
# -----------------------------------------------------------------------------


def colorize_status(status):
    """ colorize the service.status field """
    return Text(
        status,
        style={
            InventoryStatusOption.active: "bright_green",
            InventoryStatusOption.pending: "bright_yellow",
        }.get(status),
    )


def make_services_table(services: List[Dict]) -> Table:
    """
    Create a Rich.Table of service inventory records.

    Parameters
    ----------
    services: List[Dict]
        Service inventory records in the form of the API dict.

    Returns
    -------
    The Table ready for console rendering.
    """
    count = len(services)
    title = f"Services ({count})" if count > 1 else "Service"

    table = Table(
        title=Text(title, style="bright_white", justify="left"),
        show_header=True,
        header_style="bold magenta",
        show_lines=True,
    )

    table.add_column("Name")
    table.add_column("Status")
    table.add_column("Product")
    table.add_column("Circuit Id")
    table.add_column("Bandwidth")
    table.add_column("Location A")
    table.add_column("Location Z")

    def make_location(_loc):
        """ create address from location fields """
        return f"{_loc['name']}\n{_loc['city']}, {_loc['state']} {_loc['postalCode']}"

    for rec in services:
        comps = rec["components"][0]

        table.add_row(
            rec["serviceName"],
            colorize_status(rec["status"]),
            f"{rec['productGroup']}\n{rec['productCategory']}",
            comps["circuitId"],
            comps["bandwidth"],
            make_location(comps["locations"][0]),
            make_location(comps["locations"][1]),
        )

    return table


# -----------------------------------------------------------------------------
#
#                               CLI CODE BEGINS
#
# -----------------------------------------------------------------------------


@cli.group()
def svc():
    """ Inventory Service commands. """
    pass


@svc.command(name="list")
def cli_svc_inventory_list():
    """
    List service inventory.
    """
    zapi = ZayoClient()
    svc_list = zapi.get_services()
    console = Console()
    console.print(make_services_table(services=svc_list))



# Copyright 2020, Jeremy Schulman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

package_name = "pyzayo"
package_version = open("VERSION").read().strip()


def requirements(filename="requirements.txt"):
    return open(filename.strip()).readlines()


with open("README.md", "r") as fh:
    long_description = fh.read()


# -----------------------------------------------------------------------------
#
#                                 Main Setup
#
# -----------------------------------------------------------------------------

setup(
    name=package_name,
    version=package_version,
    description="Zayo API python client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jeremy Schulman",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements(),
    entry_points={"console_scripts": ["zayocli = pyzayo.cli.__main__:main"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Networking",
    ],
)



"""
This file contains the CLI code for the maintenance commands.

References
----------
    For the Rich package, colors are defined here:
    https://rich.readthedocs.io/en/latest/appendix/colors.html#appendix-colors
"""

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import List
from operator import attrgetter

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

import click
from rich.console import Console
from rich.table import Table, Text
import maya

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from pyzayo import ZayoClient
from .cli_root import cli
from pyzayo import consts
from pyzayo.mtc_models import CaseRecord, ImpactRecord, NotificationDetailRecord
from pyzayo.consts import CaseStatusOptions

# -----------------------------------------------------------------------------
#
#                               TABLE CODE BEGINS
#
# -----------------------------------------------------------------------------


def colorize_urgency(urgency: str):
    """ set the text style for case.urgency field """
    style = {
        consts.CaseUrgencyOptions.emergency: "bold red",
        consts.CaseUrgencyOptions.demand: "bright_blue",
        consts.CaseUrgencyOptions.planned: "bright_yellow",
    }.get(
        consts.CaseUrgencyOptions(urgency)  # noqa
    )  # noqa

    return Text(urgency, style=style)


def colorize_status(status):
    """ set the text style for case.status field"""
    return Text(
        status,
        style={CaseStatusOptions.scheduled: "bright_yellow"}.get(
            consts.CaseStatusOptions(status)  # noqa
        ),
    )


def colorize_impact(impact):
    """ set the text style for case.impact field """
    style = {
        consts.CaseImpactOptions.potential_svc_aff: "",
        consts.CaseImpactOptions.svc_aff: "bold red",
    }.get(
        consts.CaseImpactOptions(impact)  # noqa
    )  # noqa

    return Text("\n".join(impact.split()), style=style)


def make_cases_table(recs: List[CaseRecord]) -> Table:
    """
    This function creates the Rich.Table that contains the cases information.

    Parameters
    ----------
    recs: List[CaseRecord]
        The list of case records in model-object form.

    Returns
    -------
    The rendered Table of case information.
    """
    n_cases = len(recs)
    table = Table(
        title=Text(
            f"Cases ({n_cases})" if n_cases > 1 else "Case",
            style="bright_white",
            justify="left",
        ),
        show_header=True,
        header_style="bold magenta",
        show_lines=True,
    )

    table.add_column("Case #")
    table.add_column("Urgency")
    table.add_column("Status")
    table.add_column("Impact")
    table.add_column("Date(s)")
    table.add_column("Location", width=12, overflow="fold")
    table.add_column("Start Time")
    table.add_column("End Time")
    table.add_column("Reason")

    pdates = attrgetter("primary_date", "primary_date_2", "primary_date_3")

    for row_obj in recs:

        if row_obj.status != consts.CaseStatusOptions.closed:
            row_obj.urgency = colorize_urgency(row_obj.urgency)  # noqa
            row_obj.impact = colorize_impact(row_obj.impact)
            row_obj.status = colorize_status(row_obj.status)

        rec_pdates = sorted(pd for pd in pdates(row_obj) if pd)
        md = maya.parse(rec_pdates[0])
        dstr = "\n".join(map(str, rec_pdates)) + f"\n({md.slang_time()})"

        table.add_row(
            row_obj.case_num,
            row_obj.urgency,
            row_obj.status,
            row_obj.impact,
            dstr,
            row_obj.location,
            str(row_obj.from_time),
            str(row_obj.to_time),
            row_obj.reason,
        )

    return table


def make_impacts_table(impacts: List[dict]) -> Table:
    """
    This function creates the Rich.Table that contains the case impact information.

    Parameters
    ----------
    impacts: List[dict]
        The list of case impact records in API dict form.

    Returns
    -------
    The rendered Table of case impact information.
    """
    count = len(impacts)

    table = Table(
        title=Text(
            f"Impacts ({count})" if count > 1 else "Impact",
            style="bright_white",
            justify="left",
        ),
        show_header=True,
        header_style="bold magenta",
        show_lines=True,
    )

    table.add_column("Case #")
    table.add_column("Circuit Id")
    table.add_column("Expected Impact")
    table.add_column("CLLI A")
    table.add_column("CLLI Z")

    for rec in impacts:
        row_obj = ImpactRecord.parse_obj(rec)

        table.add_row(
            row_obj.case_num,
            row_obj.circuit_id,
            row_obj.impact,
            row_obj.clli_a,
            row_obj.clli_z,
        )

    return table


def make_notifs_table(notifs):
    """
    This function creates the Rich.Table that contains the case notification information.

    Parameters
    ----------
    notifs: List[dict]
        The list of case impact records in API dict form.

    Returns
    -------
    The rendered Table of case notifications information.
    """
    count = len(notifs)
    table = Table(
        title=Text(
            f"Notifications ({count})" if count > 1 else "Notification",
            style="bright_white",
            justify="left",
        ),
        show_header=True,
        header_style="bold magenta",
        show_lines=True,
    )

    table.add_column("#")
    table.add_column("Type")
    table.add_column("Email Sent")
    table.add_column("Email Subject")
    table.add_column("Email To")

    for rec in notifs:
        row_obj = NotificationDetailRecord.parse_obj(rec)
        email_list = sorted(map(str.strip, row_obj.email_list.split(";")))
        mt = maya.parse(row_obj.date)
        dstring = (
            mt.local_datetime().strftime("%Y-%m-%d\n%H:%M:%S")
            + f"\n({mt.slang_time()})"
        )

        table.add_row(
            row_obj.name, row_obj.type, dstring, row_obj.subject, "\n".join(email_list)
        )

    return table


# -----------------------------------------------------------------------------
#
#                               CLI CODE BEGINS
#
# -----------------------------------------------------------------------------


@cli.group()
def mtc():
    """
    Maintenance commands.
    """
    pass


@mtc.command(name="cases")
def mtc_cases():
    """
    Show listing of maintenance caess.
    """
    zapi = ZayoClient()
    recs = [
        rec
        for rec in map(
            CaseRecord.parse_obj,
            zapi.get_cases(orderBy=[consts.OrderBy.date_sooner.value]),
        )
        if rec.status != CaseStatusOptions.closed
    ]
    console = Console()
    console.print(make_cases_table(recs))


@mtc.command(name="case-details")
@click.argument("case_number")
def mtc_case_details(case_number):
    """
    Show specific case details.
    """
    zapi = ZayoClient()

    # find the case by number

    case, impacts, notifs = zapi.get_case_details(by_case_num=case_number)

    console = Console()

    if not case:
        console.print(f"Case [bold white]{case_number}: [bold red]Not found")
        return

    # TODO raw dumping data for now ... need to make pretty.

    console.print(f"\nCase [bold white]{case_number}[/bold white]: [bold green]Found")
    console.print("\n", make_cases_table([CaseRecord.parse_obj(case)]), "\n")
    console.print(make_impacts_table(impacts), "\n")
    console.print(make_notifs_table(notifs), "\n")

    # console.print(Syntax(code=json.dumps(case, indent=3), lexer_name="json"))



"""
This file contains the ZayoClientBase base class that is used when subclassing
specific API functional areas, such as maintenance.
"""
# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import Optional, List, Dict
import math
from os import getenv
import asyncio
from itertools import chain

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

import httpx
from tenacity import retry, wait_random_exponential, retry_if_exception

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from pyzayo import consts
from pyzayo.api import ZayoAPI

# -----------------------------------------------------------------------------
# Module Exports
# -----------------------------------------------------------------------------

__all__ = ["ZayoClientBase"]

# -----------------------------------------------------------------------------
#
#                               CODE BEGINS
#
# -----------------------------------------------------------------------------


class ZayoClientBase(object):
    """
    This is a base class for any Zayo Client.  This class provides the common
    functionality that would be used by subclassed clients such as the Zayo
    maintenance client, ZayoMatenanceMixin.
    """

    def __init__(self):
        """ authorize to the ZAYO API and setup for the mainteance functioanl area """
        self._auth_payload: Optional[dict] = None
        self.authenticate()
        self.api = ZayoAPI(base_url=consts.ZAYO_URL_SM, access_token=self.access_token)

    @property
    def access_token(self):
        """ returns the current access token value """
        return self._auth_payload["access_token"]

    def authenticate(self):
        """
        This method is used to authenticate to the Zayo API system using the
        client-id and client-secret values obtained from the environment.

        This method is called during instance initialization and the access
        token can be obtained via the `access_token` property.

        Notes
        -----
        According to the Zayo API documentation, a token is valid for 1hr. Plan
        accordingly.
        """
        client_id = getenv("ZAYO_CLIENT_ID")
        client_secret = getenv("ZAYO_CLIENT_SECRET")
        payload = {
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "client_credentials",
            "scope": "openid",
        }
        res = httpx.post(url=consts.ZAYO_URL_AUTH, data=payload)
        res.raise_for_status()
        self._auth_payload = res.json()

    def get_records_count(self, url, **params) -> int:
        """
        This function will return the total number of records that match the
        request criteria defined by `params`.  If `params` is not provided, then
        the count of all records for the given URL.

        Parameters
        ----------
        url: str
            The route endpoint providing recoreds, for example the
            value defined in the `consts.ZAYO_SM_ROUTE_MTC_CASES`.

        Other Parameters
        ----------------
        The other `params` are key-values as defined by the Zayo API
        spec; and these are specific to the URL used.

        Returns
        -------
        The number of records matching the criterial (or all)
        """
        loop = asyncio.get_event_loop()

        # do not request any records to be returned, just need the record count
        # from the metadata response.

        payload = params.copy()
        payload["paging"] = {"top": 0}
        res = loop.run_until_complete(self.api.post(url, json=payload))
        res.raise_for_status()
        return res.json()["data"]["metadata"]["totalRecordCount"]

    def paginate_records(self, url, **params) -> List[Dict]:
        """
        This function will return all records for a given request criterial
        determined by `params` or all records.

        Parameters
        ----------
        url: str
            The API route endpoint

        Other Parameters
        ----------------
        key-values speciifc to the url being used, determines the request
        criteria.

        Notes
        -----
        The API is limited to return a maximum of 100 records per the "top" and
        "step" fields of the paging criteria.  In the event there are more than
        100 records, only the first 100 records will be returned.  The Caller
        could use the `get_records_count` to determine the total value before
        calling this function to create a better `params` criterial (perhaps by
        date range) to avoid > 100 records.

        Returns
        -------
        List of records, each dict schema is specific to the url.
        """
        if params:
            paging = params.setdefault("paging", {})
            page_sz = paging.setdefault("top", consts.MAX_TOP_COUNT)
        else:
            page_sz = consts.MAX_TOP_COUNT
            params = dict(paging=dict(top=page_sz, skip=0))

        loop = asyncio.get_event_loop()
        total_recs = self.get_records_count(url=url, **params)

        # TODO: limit to 100?  API seems to allow to page beyond skip=50 ...
        # max_recs = min(total_recs, consts.MAX_PAGED_RECORDS)

        max_recs = total_recs
        total_pages = math.ceil(max_recs / page_sz)

        tasks = list()
        # print(f"paging total records {max_recs} pages {total_pages} of size {page_sz}")

        @retry(
            retry=retry_if_exception(httpx.ReadTimeout),
            wait=wait_random_exponential(multiplier=1, max=10),
        )
        async def get_page(payload):
            """ get a page and retry if exception """
            return await self.api.post(url, json=payload)

        for page in range(total_pages):
            task_params = params.copy()
            task_params["paging"] = {"top": page_sz, "skip": (page * page_sz)}
            tasks.append(get_page(task_params))

        http_res_list = loop.run_until_complete(
            asyncio.gather(*tasks, return_exceptions=True)
        )
        return list(
            chain.from_iterable(
                resp.json()["data"]["records"]
                for resp in http_res_list
                if resp.is_error is False
            )
        )



"""
This module contains the Zayo API client for accessing the Maintenance
endpoints.

References
----------
API documentation:
    http://54.149.224.75/wp-content/uploads/2020/03/Maintenance-Cases-Wiki.pdf
"""

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import List, Dict
import asyncio

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

from first import first

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from pyzayo.base_client import ZayoClientBase
from pyzayo import consts

# -----------------------------------------------------------------------------
# Package Exports
# -----------------------------------------------------------------------------

__all__ = ["ZayoMatenanceMixin"]


# -----------------------------------------------------------------------------
#
#                               CODE BEGINS
#
# -----------------------------------------------------------------------------


class ZayoMatenanceMixin(ZayoClientBase):
    """
    This class defines the Zayo Maintenace API client instance.

    References
    ----------
    API documentation:
    http://54.149.224.75/wp-content/uploads/2020/03/Maintenance-Cases-Wiki.pdf
    """

    # def __init__(self):
    #     """ setup client to use the Maintenace base URL """
    #     super(ZayoMatenanceMixin, self).__init__(base_url=consts.ZAYO_URL_SM)

    def get_cases(self, **params) -> List[Dict]:
        """
        Returns the maintenance cases.  If `params` are provided they are used
        as-is in the body-request per the API spec.  For example

            filter={'caseNumber': 'TNN-0003153584"}
            paging={'top': 5}

        Parameters
        ----------
        params

        Returns
        -------
        List[Dict]
        """
        return self.paginate_records(url=consts.ZAYO_SM_ROUTE_MTC_CASES, **params)

    def get_case(self, by_case_num: str) -> Dict:
        """
        This method will return the specific case record identified `by_case_num`.

        Parameters
        ----------
        by_case_num: str
            The case number

        Returns
        -------
        The record dictionary as defined by the API.
        """
        recs = self.paginate_records(
            url=consts.ZAYO_SM_ROUTE_MTC_CASES, filter={"caseNumber": by_case_num}
        )
        return first(recs)

    def get_case_details(self, by_case_num: str):
        """
        This method will obtain all of the impact and notification details
        associated with a given case number.

        Parameters
        ----------
        by_case_num: str
            The case number, starts with "TNN-"

        Returns
        -------
        Tuple [case, impacts, notifs_details]
            case: dict
                The case records

            impacts: list[dict]
                List of impact records

            notifs_details: list[dict]
                List of notificaiton detail records
        """
        case = self.get_case(by_case_num=by_case_num)

        if not case:
            return None, None, None

        impacts = self.get_impacts(by_case_num=by_case_num)

        notif_details = [
            self.get_notification_details(by_name=notif["name"])
            for notif in self.get_notifications(by_case_num=by_case_num)
        ]

        return case, impacts, notif_details

    def get_impacts(self, by_circuit_id=None, by_case_num=None, **params) -> List[Dict]:
        """
        Get the maintenance impact records.  If `by_circuid_id` or `by_case_num`
        are provided, then these are the primary request criteria.  If not provided
        then all impact records are requested subject to any additional `params`.

        Parameters
        ----------
        by_circuit_id: str
            The circuit ID to match impact records

        by_case_num: str
            The case number to match impact records


        Other Parameters
        ----------------
        Used as-is per the API spec for request matching.


        Returns
        -------
        List of impact records matching the request criteria.
        """
        if by_case_num:
            req_filter = dict(caseNumber=by_case_num)
        elif by_circuit_id:
            req_filter = dict(circuitId=by_circuit_id)
        else:
            req_filter = {}

        params = {"filter": req_filter, **params}
        return self.paginate_records(url=consts.ZAYO_SM_ROUTE_MTC_IMPACTS, **params)

    def get_notifications(self, by_case_num) -> List[Dict]:
        """
        Get notifications by case number.

        Parameters
        ----------
        by_case_num: str
            The case number to match, begins with "TTN-"

        Returns
        -------
        List of notification records.
        """
        loop = asyncio.get_event_loop()

        res = loop.run_until_complete(
            self.api.get(
                url=consts.ZAYO_SM_ROUTE_MTC_NOTIFS_BY_CASE.format(case_num=by_case_num)
            )
        )

        res.raise_for_status()
        body = res.json()
        return body["data"]

    def get_notification_details(self, by_name: str) -> Dict:
        """
        For a given ntoficication by "name" return the details dictionary.

        Parameters
        ----------
        by_name: str
            The notification name, begins with "MNN-"

        Returns
        -------
        The detail record; refer to API spec for key-value fields.
        """
        loop = asyncio.get_event_loop()

        res = loop.run_until_complete(
            self.api.get(
                url=consts.ZAYO_SM_ROUTE_MTC_NOTIFS_BY_NAME.format(name=by_name)
            )
        )

        res.raise_for_status()
        body = res.json()
        return body["data"]
